/*
 * Copyright (C) 1997-2008 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file vss.c The virtual streaming system.
 *
 * This contains the audio streaming code of para_server which is independent
 * of the current audio format, audio file selector and of the activated
 * senders.
 */

#include <sys/mman.h> /* mmap */
#include <sys/time.h> /* gettimeofday */
#include <sys/types.h>
#include <dirent.h>

#include "para.h"
#include "error.h"
#include "string.h"
#include "afh.h"
#include "afs.h"
#include "server.h"
#include "net.h"
#include "server.cmdline.h"
#include "vss.h"
#include "send.h"
#include "ipc.h"
#include "fd.h"

static struct timeval announce_tv;
static struct timeval data_send_barrier;
static struct timeval eof_barrier;
static struct timeval autoplay_barrier;

extern struct misc_meta_data *mmd;
extern struct sender senders[];

static char *map;

/**
 * Check if vss status flag \a P (playing) is set.
 *
 * \return Greater than zero if playing, zero otherwise.
 *
 */
unsigned int vss_playing(void)
{
	return mmd->new_vss_status_flags & VSS_PLAYING;
}

/**
 * Check if the \a N (next) status flag is set.
 *
 * \return Greater than zero if set, zero if not.
 *
 */
unsigned int vss_next(void)
{
	return mmd->new_vss_status_flags & VSS_NEXT;
}

/**
 * Check if a reposition request is pending.
 *
 * \return Greater than zero if true, zero otherwise.
 *
 */
unsigned int vss_repos(void)
{
	return mmd->new_vss_status_flags & VSS_REPOS;
}

/**
 * Check if the vss is currently paused.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_paused(void)
{
	return !(mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

/**
 * Check if the vss is currently stopped.
 *
 * \return Greater than zero if paused, zero otherwise.
 *
 */
unsigned int vss_stopped(void)
{
	return (mmd->new_vss_status_flags & VSS_NEXT)
		&& !(mmd->new_vss_status_flags & VSS_PLAYING);
}

/**
 * Initialize the virtual streaming system.
 *
 * This also initializes all supported senders and starts streaming
 * if the --autoplay command line flag was given.
 */
void vss_init(void)
{
	int i;
	char *hn = para_hostname(), *home = para_homedir();
	long unsigned announce_time = conf.announce_time_arg > 0?
			conf.announce_time_arg : 300,
		autoplay_delay = conf.autoplay_delay_arg > 0?
			conf.autoplay_delay_arg : 0;
	ms2tv(announce_time, &announce_tv);
	PARA_INFO_LOG("announce timeval: %lums\n", tv2ms(&announce_tv));
	for (i = 0; senders[i].name; i++) {
		PARA_NOTICE_LOG("initializing %s sender\n", senders[i].name);
		senders[i].init(&senders[i]);
	}
	free(hn);
	free(home);
	if (conf.autoplay_given) {
		struct timeval now, tmp;
		mmd->vss_status_flags |= VSS_PLAYING;
		mmd->new_vss_status_flags |= VSS_PLAYING;
		gettimeofday(&now, NULL);
		ms2tv(autoplay_delay, &tmp);
		tv_add(&now, &tmp, &autoplay_barrier);
	}
}

static int chk_barrier(const char *bname, const struct timeval *now,
		const struct timeval *barrier, struct timeval *diff,
		int print_log)
{
	long ms;

	if (tv_diff(now, barrier, diff) > 0)
		return 1;
	ms = tv2ms(diff);
	if (print_log && ms)
		PARA_DEBUG_LOG("%s barrier: %lims left\n", bname, ms);
	return -1;
}

static void vss_next_chunk_time(struct timeval *due)
{
	struct timeval tmp;

	tv_scale(mmd->chunks_sent, &mmd->afd.afhi.chunk_tv, &tmp);
	tv_add(&tmp, &mmd->stream_start, due);
}

/*
 * != NULL: timeout for next chunk
 * NULL: nothing to do
 */
static struct timeval *vss_compute_timeout(void)
{
	static struct timeval the_timeout;
	struct timeval now, next_chunk;

	if (vss_next() && map) {
		/* only sleep a bit, nec*/
		the_timeout.tv_sec = 0;
		the_timeout.tv_usec = 100;
		return &the_timeout;
	}
	gettimeofday(&now, NULL);
	if (chk_barrier("autoplay_delay", &now, &autoplay_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("eof", &now, &eof_barrier, &the_timeout, 1) < 0)
		return &the_timeout;
	if (chk_barrier("data send", &now, &data_send_barrier,
			&the_timeout, 1) < 0)
		return &the_timeout;
	if (!vss_playing() || !map)
		return NULL;
	vss_next_chunk_time(&next_chunk);
	if (chk_barrier("chunk", &now, &next_chunk, &the_timeout, 0) < 0)
		return &the_timeout;
	/* chunk is due or bof */
	the_timeout.tv_sec = 0;
	the_timeout.tv_usec = 0;
	return &the_timeout;
}

static void vss_eof(void)
{
	struct timeval now;
	int i;
	char *tmp;

	if (!map) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		return;
	}
	if (mmd->new_vss_status_flags & VSS_NOMORE)
		mmd->new_vss_status_flags = VSS_NEXT;
	gettimeofday(&now, NULL);
	tv_add(&mmd->afd.afhi.eof_tv, &now, &eof_barrier);
	munmap(map, mmd->size);
	map = NULL;
	mmd->chunks_sent = 0;
	mmd->offset = 0;
	mmd->afd.afhi.seconds_total = 0;
	free(mmd->afd.afhi.chunk_table);
	mmd->afd.afhi.chunk_table = NULL;
	tmp  = make_message("%s:\n%s:\n%s:\n", status_item_list[SI_AUDIO_FILE_INFO],
		status_item_list[SI_TAGINFO1], status_item_list[SI_TAGINFO2]);
	strncpy(mmd->afd.afhi.info_string, tmp, sizeof(mmd->afd.afhi.info_string));
	mmd->afd.afhi.info_string[sizeof(mmd->afd.afhi.info_string) - 1] = '\0';
	make_empty_status_items(mmd->afd.verbose_ls_output);
	free(tmp);
	mmd->mtime = 0;
	mmd->size = 0;
	mmd->events++;
}

/**
 * Get the header of the current audio file.
 *
 * \param header_len The length of the header is stored here.
 *
 * \return A pointer to a buffer containing the header, or NULL, if no audio
 * file is selected or if the current audio format does not need special header
 * treamtment.
 *
 */
char *vss_get_header(size_t *header_len)
{
	if (!map || !mmd->afd.afhi.header_len)
		return NULL;
	*header_len = mmd->afd.afhi.header_len;
	return map + mmd->afd.afhi.header_offset;
}

/**
 * Get the list of all supported audio formats.
 *
 * \return Aa space separated list of all supported audio formats
 * It is not allocated at runtime, i.e. there is no need to free
 * the returned string in the caller.
 */
const char *supported_audio_formats(void)
{
	return SUPPORTED_AUDIO_FORMATS;
}

/**
 * Get the chunk time of the current audio file.
 *
 * \return A pointer to a struct containing the chunk time, or NULL,
 * if currently no audio file is selected.
 */
struct timeval *vss_chunk_time(void)
{
	if (!map)
		return NULL;
	return &mmd->afd.afhi.chunk_tv;
}

/** The possible states of the afs socket. See \ref afs_socket. */
enum afs_socket_status {
	/** Socket is inactive. */
	AFS_SOCKET_READY,
	/** Socket fd was included in the write fd set for select(). */
	AFS_SOCKET_CHECK_FOR_WRITE,
	/** vss wrote a request to the socket and waits for afs to reply. */
	AFS_SOCKET_AFD_PENDING
};

static enum afs_socket_status afsss;

/**
 * Compute the timeout for para_server's main select-loop.
 *
 * This function gets called from para_server to determine the timeout value
 * for its main select loop.
 *
 * \param rfds The set of file descriptors to be checked for reading.
 * \param wfds The set of file descriptors to be checked for writing.
 * \param max_fileno The highest-numbered file descriptor.
 *
 * Before the timeout is computed, the current vss status flags are evaluated
 * and acted upon by calling appropriate functions from the lower layers.
 * Possible actions include
 *
 * 	- request a new audio file from afs,
 *	- shutdown of all senders (stop/pause command),
 *	- reposition the stream (ff/jmp command).
 *
 * \return A pointer to a struct timeval containing the timeout for the next
 * chunk of data to be sent, or NULL if we're not sending right now.
 */
struct timeval *vss_preselect(fd_set *rfds, fd_set *wfds, int *max_fileno)
{
	int i;
	struct timeval *tv;

	para_fd_set(afs_socket, rfds, max_fileno);
	if (!map)
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
	else {
		if (vss_next()) {
			vss_eof();
			return vss_compute_timeout();
		}
	}
	if (vss_paused() || vss_repos()) {
		for (i = 0; senders[i].name; i++)
			senders[i].shutdown_clients();
		if (map) {
			struct timeval now;
			gettimeofday(&now, NULL);
			if (!vss_paused() || mmd->chunks_sent)
				tv_add(&mmd->afd.afhi.eof_tv, &now, &eof_barrier);
			if (vss_repos())
				tv_add(&now, &announce_tv, &data_send_barrier);
		}
		mmd->chunks_sent = 0;
	}
	if (vss_repos()) {
		mmd->new_vss_status_flags &= ~(VSS_REPOS);
		mmd->current_chunk = mmd->repos_request;
	}
	tv = vss_compute_timeout();
	if (tv)
		return tv;
	if (!map && vss_playing() &&
			!(mmd->new_vss_status_flags & VSS_NOMORE)) {
		if (afsss == AFS_SOCKET_READY ||
				afsss == AFS_SOCKET_CHECK_FOR_WRITE) {
			PARA_DEBUG_LOG("ready and playing, but no audio file\n");
			para_fd_set(afs_socket, wfds, max_fileno);
			afsss = AFS_SOCKET_CHECK_FOR_WRITE;
		}
	}
	return tv;
}

static int recv_afs_msg(int *fd, uint32_t *code, uint32_t *data)
{
	char control[255], buf[8];
	struct msghdr msg = {.msg_iov = NULL};
	struct cmsghdr *cmsg;
	struct iovec iov;
	int ret = 0;

	*fd = -1;
	iov.iov_base = buf;
	iov.iov_len = sizeof(buf);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	memset(buf, 0, sizeof(buf));
	ret = recvmsg(afs_socket, &msg, 0);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	afsss = AFS_SOCKET_READY;
	if (iov.iov_len != sizeof(buf))
		return -E_SHORT_AFS_READ;
	*code = *(uint32_t*)buf;
	*data =  *(uint32_t*)(buf + 4);
	for (cmsg = CMSG_FIRSTHDR(&msg); cmsg; cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level != SOL_SOCKET
			|| cmsg->cmsg_type != SCM_RIGHTS)
		continue;
		if ((cmsg->cmsg_len - CMSG_LEN(0)) / sizeof(int) != 1)
		continue;
		*fd = *(int *)CMSG_DATA(cmsg);
	}
	return 1;
}

static void recv_afs_result(void)
{
	int ret, passed_fd, shmid;
	uint32_t afs_code = 0, afs_data = 0;
	struct stat statbuf;
	struct timeval now;

	ret = recv_afs_msg(&passed_fd, &afs_code, &afs_data);
	if (ret < 0)
		goto err;
	PARA_DEBUG_LOG("fd: %d, code: %u, shmid: %u\n", passed_fd, afs_code,
		afs_data);
	ret = -E_NOFD;
	if (afs_code != NEXT_AUDIO_FILE)
		goto err;
	if (passed_fd < 0)
		goto err;
	shmid = afs_data;
	ret = load_afd(shmid, &mmd->afd);
	if (ret < 0)
		goto err;
	shm_destroy(shmid);
	ret = fstat(passed_fd, &statbuf);
	if (ret < 0) {
		PARA_ERROR_LOG("fstat error:\n");
		ret = -ERRNO_TO_PARA_ERROR(errno);
		goto err;
	}
	mmd->size = statbuf.st_size;
	mmd->mtime = statbuf.st_mtime;
	map = para_mmap(mmd->size, PROT_READ, MAP_PRIVATE,
		passed_fd, 0);
	close(passed_fd);
	mmd->chunks_sent = 0;
	mmd->current_chunk = 0;
	mmd->offset = 0;
	mmd->events++;
	mmd->num_played++;
	mmd->new_vss_status_flags &= (~VSS_NEXT);
	gettimeofday(&now, NULL);
	tv_add(&now, &announce_tv, &data_send_barrier);
	return;
err:
	if (passed_fd >= 0)
		close(passed_fd);
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	mmd->new_vss_status_flags = VSS_NEXT;
}

void vss_post_select(fd_set *rfds, fd_set *wfds)
{
	int ret;

	if (FD_ISSET(afs_socket, rfds))
		recv_afs_result();
	if (afsss != AFS_SOCKET_CHECK_FOR_WRITE || !FD_ISSET(afs_socket, wfds))
		return;
	PARA_NOTICE_LOG("requesting new fd from afs\n");
	ret = send_buffer(afs_socket, "new");
	afsss = AFS_SOCKET_AFD_PENDING;
}

static void get_chunk(long unsigned chunk_num, char **buf, size_t *len)
{
	size_t pos = mmd->afd.afhi.chunk_table[chunk_num];
	*buf = map + pos;
	*len = mmd->afd.afhi.chunk_table[chunk_num + 1] - pos;
}

/**
 * Get the data of the given chunk.
 *
 * \param chunk_num The number of the desired chunk.
 * \param buf Chunk data.
 * \param len Chunk length in bytes.
 *
 * \return Standard.
 */
int vss_get_chunk(long unsigned chunk_num, char **buf, size_t *len)
{
	if (!map || !vss_playing())
		return -E_CHUNK;
	if (chunk_num >= mmd->afd.afhi.chunks_total)
		return -E_CHUNK;
	get_chunk(chunk_num, buf, len);
	return 1;
}

/**
 * Main sending function.
 *
 * This function gets called from para_server as soon as the next chunk of data
 * should be pushed out. It obtains a pointer to the data to be sent out as
 * well as its length from mmd->afd.afhi. This information is then passed to
 * each supported sender's send() function which is supposed to send out the data
 * to all connected clients.
 */
void vss_send_chunk(void)
{
	int i;
	struct timeval now, due;
	char *buf;
	size_t len;

	if (!map || !vss_playing())
		return;
	gettimeofday(&now, NULL);
	vss_next_chunk_time(&due);
	if (tv_diff(&due, &now, NULL) > 0)
		return;
	if (chk_barrier("eof", &now, &eof_barrier, &due, 1) < 0)
		return;
	if (chk_barrier("data send", &now, &data_send_barrier,
			&due, 1) < 0)
		return;
	mmd->new_vss_status_flags &= ~VSS_REPOS;
	if (mmd->current_chunk >= mmd->afd.afhi.chunks_total) { /* eof */
		mmd->new_vss_status_flags |= VSS_NEXT;
		return vss_eof();
	}
	/*
	 * We call the send function also in case of empty chunks as they
  	 * might have still some data queued which can be sent in this case.
  	 */
	if (!mmd->chunks_sent) {
		struct timeval tmp;
		gettimeofday(&mmd->stream_start, NULL);
		tv_scale(mmd->current_chunk, &mmd->afd.afhi.chunk_tv, &tmp);
		mmd->offset = tv2ms(&tmp);
		mmd->events++;
	}
	get_chunk(mmd->current_chunk, &buf, &len);
	for (i = 0; senders[i].name; i++)
		senders[i].send(mmd->current_chunk, mmd->chunks_sent, buf, len);
	mmd->new_vss_status_flags |= VSS_PLAYING;
	mmd->chunks_sent++;
	mmd->current_chunk++;
}
